//This is free and unencumbered software is provided by Cyrstalfontz as an example of
//how to program and use their line of ePaper displays. 
//
//For more information on these displays please visit: http://Crystalfontz.com
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non - commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain.We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors.We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>

#include <SPI.h>


#define CMD  false
#define DATA true

#define SET_PANEL   0x00
#define SET_PWR_REG 0x01
#define PWR_OFF     0x02
#define PWR_ON      0x04
#define BTST        0x06
#define REFRESH_CMD 0x12
#define SET_PLL     0x30
#define TEMP_SENSE  0x41
#define SET_VCOM    0x50
#define SET_RES     0x61
#define VCOM_DC     0x82
#define DATA_STOP   0x11
#define DATA_START1 0x10
#define DATA_START2 0x13

#define cmd_00_default 0xc3
#define cmd_30_default 0x2a
#define cmd_41_default 0x00
#define cmd_50_default 0x27
#define cmd_82_default 0x0a
#define vcom_fix 0x0a

static unsigned char cmd_01_data[] = {0x07,0x00,0x0a,0x00};
static unsigned char cmd_06_data[] = {0x07,0x06,0x05};
static unsigned char cmd_61_data[] = {0x68,0x00,0xd4};

class ePaper {
  private:
  int BUSY;
  int RESET;
  int CS;
  int DC;
  int BUSSEL;
  
public:
	ePaper(){}
 	~ePaper(){}
  void config_Pins(int Busy_Pin, int Reset_Pin, int CS_Pin,int DC_Pin, int BusSel_Pin);
  void WriteByte(unsigned char Byte, bool data);
  void WriteDataBlock(unsigned char *data, int length);
  void WriteFlashImage(unsigned char cmd,const unsigned char *data, int length,bool inverted);
  void SendCMD(unsigned char cmd,unsigned char * data, int length);
  void DownloadImage(unsigned char cmd,unsigned char * data, int length);
  void Reset(void);
  void LoadLUTs(void);
 
  bool CheckStatus(void);

  inline void PowerOff(void)  {WriteByte(PWR_OFF,CMD);}
  inline void PowerOn(void)   {WriteByte(PWR_ON,CMD);}
  inline void Refresh(void)   {WriteByte(REFRESH_CMD,CMD);}
  inline void DataStop(void)  {WriteByte(DATA_STOP,CMD);}
  
inline bool SetPanel(unsigned char Byte = cmd_00_default)
  {
    bool retcode;
    if(retcode = CheckStatus())
    {
      WriteByte(SET_PANEL,CMD);
      WriteByte(Byte,DATA);
    }
    return retcode;
  }
// Most of the following member functions consist of defaults which are known to 
// work with this display. They may be overidden to provide additional functionality.
  inline void SetPwrReg(unsigned char data[] = cmd_01_data, int length = sizeof(cmd_01_data))
  {
    SendCMD(SET_PWR_REG,data,length);
  }

 inline void BoostSoftStart(unsigned char data[] = cmd_06_data, int length = sizeof(cmd_06_data))
  {
    SendCMD(BTST,data,length);
  }

inline void Set_PLL(unsigned char Byte = cmd_30_default)
  {   
    WriteByte(SET_PLL,CMD);
    WriteByte(Byte,DATA);
  }
inline void Set_VCOM(unsigned char Byte = cmd_50_default)
  {
     WriteByte(SET_VCOM,CMD);
     WriteByte(Byte,DATA);
  }
  
inline void Set_TempSense(unsigned char Byte = cmd_41_default)
  {
     WriteByte(TEMP_SENSE,CMD);
     WriteByte(Byte,DATA);
  }
inline void Set_Resolution(unsigned char data[] = cmd_61_data, int length = sizeof(cmd_61_data))
  {
    SendCMD(SET_RES,data,length);
  }
inline void Set_VCOM_DC(unsigned char Byte = vcom_fix)
  {
     WriteByte(VCOM_DC,CMD);
     WriteByte(Byte,DATA);
  }


};

